/**
 * @author Jon Rune Paulsen
 * Version: 13.03.2007
 * Modifications from last time:
 * 		- flyttet k-means til egen klasse
 * 		- opprettet klassene HierarchicalClustering og NyVersjonAvkmeans
 * 		- la til funksjoner for de tre metodene
 * 		- kommenterte ut funksjonene til k-means
 */


package no.omim.engine;

import org.apache.log4j.*;

import org.apache.lucene.index.IndexReader;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.index.TermDocs;
import org.apache.lucene.index.TermEnum;
//import org.apache.lucene.index.TermFreqVector;
import org.apache.lucene.store.Directory;
import org.apache.lucene.store.FSDirectory; // opprinnelig RAMDirectory
//import org.apache.lucene.store.RAMDirectory;
import org.apache.lucene.analysis.Analyzer;

import Jama.*; // For  foreta SVD.

import no.omim.engine.analyzer.OMIMAnalyzer;
import no.omim.engine.analyzer.StopWords;

//import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
//import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;

// End import

public class Garcia {
	
	// variables	
	private Directory _ramDirectory; // opprinnelig RAMDirectory
	private Analyzer _analyzer;
	private IndexWriter _ramWriter;
	private IndexReader _ramReader;
	
	private Matrix _X;
	private Matrix _U;
	private Matrix _S;
	private Matrix _V;
	private Matrix _Uk; // brukes til  konvertere query til document-concept space
	private Matrix _Sk;
	private Matrix _Vk;
	private Matrix _Skinv; // brukes til  konvertere query til document-concept space
	
	private int _numTerms;
	private int _numDocs;
	
	private int _k = 10; // FLGENDE VERDI SKAL BRUKES P ENDELIG VERSJON: 100
	
	//private int _numClusters = 5; // antall klynger.
	//private ClusterList _arrClusters;
	
	private ArrayList<String> _documents; // TODO sjekk
	//String _cluster[][];
	//double _clusterVector[];
	
	
	private static Logger logger = Logger.getLogger("applogger");
	
	/*
	 * Constructor.
	 * @param strIndexDirectory The path to store the index
	 */
	public Garcia(String strIndexDirectory) throws IOException {
		// utkommenteringen nedenfor er kun for  f kjrt dette uten  mtte indeksere p nytt!
		this._ramDirectory = FSDirectory.getDirectory(strIndexDirectory, true); // opprinnelig new RAMDirectory();
		this._analyzer = new OMIMAnalyzer(StopWords.stopTable);
		this._ramWriter = new IndexWriter(this._ramDirectory, this._analyzer, true);
		this._numTerms = 0;
		this._numDocs = 0;
	}
	
	
	/*
	 * Set up the environment.
	 * @param fileDirectory The directory to index
	 */
	public void setUp(String fileDirectory) throws IOException {
		try
		{
			//logger.debug("File directory: " + fileDirectory);
			File testData = new File(fileDirectory);
			File[] contents = testData.listFiles();
			//logger.debug("Number of files: " + contents.length);
			if (contents != null) {
				for (int i = 0; i < contents.length; i++) {					
					this._ramWriter.addDocument(FileDocument.Document(contents[i]));
					//logger.debug("Document: " + i);
					System.out.println("Dokument: " + i);
					this._ramWriter.optimize();
					this._ramWriter.close();
				}
				//logger.debug("FERDIG");
				System.out.println("FERDIG");
			}
		}
		catch (IOException ioe) {
			//logger.error(ioe);
			throw ioe;
		}		
	}
	
	/*
	 * Transfer to SVD matrix.
	 */
	public void transferToSVDMatrix() throws Exception {
		try {
			// TODO f den til  lese fra indeksfil!
			this._ramReader = IndexReader.open("C:/indexDirectory/luceneIndex/");// opprinnelig this._ramDirectory
			System.out.println("pnet index");
			TermEnum enumTermer = this._ramReader.terms();
			this._numDocs = this._ramReader.numDocs();
			
			//logger.debug("transferToSVDMatrix");
			
			//TermFreqVector[] termFreqVct;
			double normFreq;
			double normWeight;
			double idf;
			int[] arr_ni;
			double maxFreq = 0;
			//int termFreqInDoc = 0;
			//String strTerm;
			String[] arr_termer; // TODO m gjres global og mulighet til  lagre til fil!
			
			// finn antall termer
			while (enumTermer.next() && enumTermer.term().field() == "contents")
			{
				this._numTerms += 1;
			}
			
			//logger.debug("this._numTerms: " + this._numTerms);
			
			// opprett array over ni
			arr_ni = new int[this._numTerms];
			
			arr_termer = new String[this._numTerms];
			
			// hent ut ni for alle termer
			enumTermer = this._ramReader.terms();
			
			//logger.debug("enumTermer: " + enumTermer);
			//logger.debug("");
			
			BufferedWriter termWriter;
			File termDestDirectory;
			
			// skriver termene til fil
			termDestDirectory = new File("C:/indexDirectory/luceneIndex/indexInfo/termOverview" + ".txt");
			termWriter = new BufferedWriter(new FileWriter(termDestDirectory));
			
			// TEST
			System.out.println("oppretter matrise");
			
			// opprett SVD-matrise
			this._X = new Matrix(this._numTerms, this._numDocs);
			System.out.println("matrise opprettet");
			//logger.debug("Strrelsen p matrisen this._X");
			//logger.debug("NumTerms: " + this._numTerms);
			//logger.debug("NumDocs: " + this._numDocs);
			//logger.debug("");
			
			TermDocs posDoc;
			// END TEST
			// lag en matrise med x antall termer og y antall dokumenter
			// for hver term, legg frekvens inn i cellen for de aktuelle dokumentene
			for (int termPos = 0; termPos < this._numTerms; termPos++) {
				System.out.println("termPos: " + termPos);
				enumTermer.next();
				// arr_ni[x]: antall dokument termen forekommer i
				arr_ni[termPos] = enumTermer.docFreq();
				//logger.debug("Termnummer: " + termPos);
				//logger.debug("Termen forekommer i " + arr_ni[termPos] + " antall dokumenter");
				arr_termer[termPos] = enumTermer.term().text();
				//logger.debug("Term: " + arr_termer[termPos]);
				//logger.debug("Field: " + enumTermer.term().field().toString());
				posDoc = this._ramReader.termDocs(enumTermer.term());
				
				while (posDoc.next()) {
					//logger.debug("Term/dokument-frekvens: " + posDoc.freq());
					//logger.debug("Term/dokument-dokument: " + posDoc.doc());
					this._X.set(termPos, posDoc.doc(), posDoc.freq());
					//logger.debug("Termfrekvens er lagt til: termPos(" + termPos +
							//"), posDoc.doc(" + posDoc.doc() + "), posDoc.freq(" +
							//posDoc.freq() + ")");
					//logger.debug("");
				}
				
				// lagrer term i fil
				termWriter.write(enumTermer.term().text()); // TODO SJEKK
				
				// hvis termPos = siste term, ikke skriv newLine til fil
				if (termPos != (this._numTerms - 1)) {
					termWriter.newLine();
				}
			}
			/*while (enumTermer.next() && enumTermer.term().field() == "contents")
			{
				arr_ni[k] = enumTermer.docFreq();
				logger.debug("enumTermer.docFreq: " + arr_ni[k]);
				arr_termer[k] = enumTermer.term().text();
				logger.debug("enumTermer.term().text()" + arr_termer[k]);
				logger.debug("Field: " + enumTermer.term().field().toString());
				posDoc = this._ramReader.termDocs(enumTermer.term());
				
				while (posDoc.next()) {
					logger.debug("TermDocs.freq: " + posDoc.freq());
					logger.debug("TermDoces.doc: " + posDoc.doc());
					this._X.set(termPos, posDoc.doc(), posDoc.freq());
					logger.debug("Termfrekvens er lagt til: termPos(" + termPos +
							"), posDoc.doc(" + posDoc.doc() + "), posDoc.freq(" +
							posDoc.freq() + ")");
					logger.debug("");
				}
				termPos++;
				
				// lagrer term i fil
				termWriter.write(enumTermer.term().text()); // TODO SJEKK
				termWriter.newLine();
				
				k += 1;
			}*/
			
			termWriter.close();
			
			// skriver dokumentene til fil
			this._documents = new ArrayList<String>();
			
			BufferedWriter docWriter;
			String docInfo;
			File docDestDirectory;
			
			docDestDirectory = new File("C:/indexDirectory/luceneIndex/indexInfo/docOverview" + ".txt");
			docWriter = new BufferedWriter(new FileWriter(docDestDirectory));
			
			//logger.debug(this._numDocs);
			
			// for alle dokumenter i index
			for (int i = 0; i < this._numDocs; i++) {
				// returns the stored fields of the ith document in this index.
				// lagrer dokumentstien i en fil
				docInfo = this._ramReader.document(i).get("path");
				
				//logger.debug(docInfo);
				
				docWriter.write(docInfo);
				// hvis det er siste dokument, ikke skriv newLine
				if (i != (this._numDocs - 1)) {
					docWriter.newLine();
				}
				
				// returns the "path" field for the document.
				this._documents.add(docInfo);
				// TODO BLIR IKKE DETTE RIKTIG DA?
				
				// termFreqVct gir NullPointerException
				//logger.debug("Lagt til dokument til Lucene-indeks");
				
				// http://www.docjar.com/html/api/org/apache/lucene/index/SegmentReader.java.html
				// http://www.docjar.com/docs/api/org/apache/lucene/index/SegmentReader.html#termVectorsReader
				
				/*TermFreqVector test;
				logger.debug("test opprettet");
				test = this._ramReader.getTermFreqVector(i, "contents");
				
				if (test == null) {
					logger.debug("test = null");
				}
				
				logger.debug("henter termFreqVector fra contents");
				int[] vectors;
				logger.debug(test.getTermFrequencies().length);
				logger.debug("test");*/
				//logger.debug("Vectors: " + Integer.toString(vectors[0]));
				
				// TODO GJR OM TIL ARRAY AV INTEGER
				// m bruke for-slyfe hvor get(int i, int j) brukes for  hente ut
				// verdiene i vektoren
				//termFreqVct = this._ramReader.getTermFreqVectors(i);
				
				//FORANDRING AV TERMFREQVCT
				double[] termFreqVec;
				termFreqVec = new double[this._numTerms];
				
				// for hver term, hent ut frekvens
				for (int term = 0; term < this._numTerms; term++) {
					termFreqVec[term] = this._X.get(term, i);
				}
				
				// END FORANDRING AV TERMFREQVCT
				
				// benytte termDocs i stedet for?
				//TermDocs test;
				//test = this._ramReader.termDocs();
				//logger.debug("TermFreqVec");
				//logger.debug(test.freq());
				//logger.debug("Dokument: " + i);
				// finn maksfrekvens i dokumentet
				//maxFreq = findMaxfrequency(termFreqVct);
				maxFreq = findMaxfrequency(termFreqVec);
				
				//logger.debug("Max Freq");
				
				// for alle termer i index
				for (int j = 0; j < arr_ni.length; j++) {
					//strTerm = arr_termer[j];
					normFreq = 0.0;
					
					//logger.debug("Finn alle termer i index");
					
					// finn termens frekvens i dokumentet
					//termFreqInDoc = findTermFrequencyInDocument(strTerm, termFreqVct);
					
					if (maxFreq > 0)
					{
						normFreq = this._X.get(j, i) / maxFreq;
						// tidligere: termFreqInDoc / maxFreq;
						//logger.debug("_X.get(j, i): " + this._X.get(j, i));
						//logger.debug("maxFreq: " + maxFreq);
						//logger.debug("Normalisert frekvens: " + normFreq);
					}
					//logger.debug("arr_ni[j]: " + arr_ni[j]);
					//logger.debug("this._numDocs: " + this._numDocs);
					// Math.log er ln (naturlig logaritme)
					idf = Math.log(this._numDocs / arr_ni[j]);
					//logger.debug("IDF: " + idf);
					normWeight = normFreq * idf;
					//logger.debug("Normalisert vekt: " + normWeight);
					
					// set(rad, kolonne, verdi)
					this._X.set(j, i, normWeight);
					//logger.debug("normWeight lagt til");
					//logger.debug("");
				}
			}
			
			docWriter.close();
			//logger.debug("TRANSFERTOSVDMATRIX FERDIG");
		}
		catch (Exception ex) {
			//logger.error(ex);
			throw ex;
		}
	}
	
	
	/*
	 * Finds the max frequency of the term.
	 * @param termFreqVct The term vector to find max frequency of
	 */
	private double findMaxfrequency(double[] termFreqVct) {
		//logger.debug("FindMaxFrequency");
		double max = 0;
		
		for (int i = 0; i < termFreqVct.length; i++) {
			//logger.debug("Term Freq: " + termFreqVct[i]);
			if (termFreqVct[i] > max)
			{
				max = termFreqVct[i];
			}
		}
		//logger.debug("Max: " + max);
		//logger.debug("");
		return max;
	}
	
	
	/*
	 * Finds the max frequency of the term.
	 * @param termFreqVct The term vector to find max frequency of
	 
	private int findMaxfrequency(TermFreqVector[] termFreqVct) throws Exception
	{
		try
		{
			logger.debug("FindMaxFrequency");
			int max = 0;
			int j;
			int k;
			int[] termFreqs;
			
			logger.debug("Length, den kommer ikke lengre enn hit");
			
			for (j = 0; j < termFreqVct.length; j++)
			{
				logger.debug("Etter termFreqVct.length;");
				termFreqs = termFreqVct[j].getTermFrequencies();
				
				for (k = 0; k < termFreqs.length; k++)
				{
					logger.debug("Term Freq: " + termFreqs[k]);
					if (termFreqs[k] > max)
					{
						max = termFreqs[k];
					}
				}
			}
			logger.debug("Max: " + Integer.toString(max));
			return max;
		}
		catch (Exception ex)
		{
			throw ex;
		}
	}
	
	
	
	
	 * Finds the term frequency in document.
	 * @param strTerm The term to find in documents
	 * @param termFreqVct The term vector to search in
	 * @return Frequency
	 
	private int findTermFrequencyInDocument(String strTerm, TermFreqVector[] termFreqVct)
	{
		int freq = 0;
		int i;
		int j;
		int[] arrFrequencies;
		String[] arrTerms;
		String strTermTemp;
		TermFreqVector termFreqVector;
		
		// alle vektorer
		for (i = 0; i < termFreqVct.length; i++)
		{
			termFreqVector = termFreqVct[i];
			arrTerms = termFreqVector.getTerms();
			arrFrequencies = termFreqVector.getTermFrequencies();
			
			// alle termer i vektor
			for (j = 0; j < arrTerms.length; j++)
			{
				strTermTemp = arrTerms[j];
				logger.debug(arrTerms[j] + " - " + arrFrequencies[j]);
				
				if (strTermTemp.equals(strTerm))
				{
					if (arrFrequencies[j] > freq)
					{
						freq = arrFrequencies[j];
					}
				}
			}		
			
		}
		
		return freq;
	}*/
	
	
	/*
	 * Perform Singular Value Decomposition and reconstruction of the matrix.
	 */
	public void performSVD() throws Exception {
		//logger.debug("performMatrixReduction");
		System.out.println("performMatrixReduction");
		performMatrixReduction();
		//logger.debug("performRankK");
		System.out.println("performRankK");
		performRankK();
		//logger.debug("writeSVDToFile");
		System.out.println("writeSVDToFile");
		writeSVDToFile();
	}
	
	
	
	/*
	 * Perform matrix reduction.
	 */
	private void performMatrixReduction() throws Exception {
		SingularValueDecomposition SVD = new SingularValueDecomposition(this._X);
		//this._U = new Matrix(this._numTerms, this._numDocs).svd().getU();
		
		System.out.println("performMatrixReduction1");
		this._U = SVD.getU();
		//logger.debug("this._U radstrrelse: " + this._U.getRowDimension());
		//logger.debug("this._U kolonnestrrelse: " + this._U.getColumnDimension());
		System.out.println("SVD getU");
		this._S = SVD.getS();
		//logger.debug("this._S radstrrelse: " + this._S.getRowDimension());
		//logger.debug("this._S kolonnestrrelse: " + this._S.getColumnDimension());
		System.out.println("SVD getS");
		this._V = SVD.getV();
		//logger.debug("this._V radstrrelse: " + this._V.getRowDimension());
		//logger.debug("this._V kolonnestrrelse: " + this._V.getColumnDimension());
		System.out.println("SVD getV");
		
		System.out.println("performMatrixReduction END");
	}
	
	
	
	/*
	 * Perform rank k on matrix U, S, V.
	 */
	private void performRankK() throws Exception {
		System.out.println("performRankK");
		// Rank k p _U.
		//logger.debug("Inni performRankK-funksjonen");
		//logger.debug("this._Uk");
		//this._Uk = new Matrix(this._numTerms, this._k);
		this._Uk = this._U.getMatrix(0, (this._numTerms - 1), 0, (this._k - 1));
		this._U = null;
		//logger.debug("this._Sk");
		this._Sk = this._S.getMatrix(0, (this._k - 1), 0, (this._k - 1));
		this._S = null;
		//logger.debug("this._Vk");
		this._Vk = this._V.getMatrix(0, (this._numDocs - 1), 0, (this._k - 1));
		this._V = null;
		//logger.debug("this.Skinv");
		this._Skinv = this._Sk.inverse(); // S invertert
		this._Sk = null;
		//logger.debug("performRankK FERDIG");
		System.out.println("performRankK END");
	}
	
	
	/*
	 * Write SVD matrices to files. Common for all of the clustering methods.
	 * Disse to matrisene blir "oppslags"-filer i forbindelse med sking, derfor
	 * lagres ikke vk-matrisen. Vk-matrisen er det som lagt i klynger av klynge-
	 * algoritmene.
	 */
	private void writeSVDToFile() throws Exception {
		System.out.println("writeSVDToFile");
		File uk;
		File skinv;
		File vk;
		String dirPath = "C:/indexDirectory/luceneIndex/SVD";
		String tmpLine;
		BufferedWriter ukOut;
		BufferedWriter skinvOut;
		BufferedWriter vkOut;
		//int ukSize = this._Uk.getRowDimension(); // antall rader (numTerms)
		//int skinvSize = this._k; // antall rader (100)
		
		uk = new File(dirPath + "/uk.txt");
		skinv = new File(dirPath + "/skinv.txt");
		vk = new File(dirPath + "/vk.txt");
		
		ukOut = new BufferedWriter(new FileWriter(uk));
		skinvOut = new BufferedWriter(new FileWriter(skinv));
		vkOut = new BufferedWriter(new FileWriter(vk));
		
		// skriv uk til fil
		for (int i = 0; i < this._numTerms; i++) { // rad
			tmpLine = "";
			for (int j = 0; j < this._k; j++) { // kolonne
				tmpLine += Double.toString(this._Uk.get(i, j)) + " "; // legger til verdi
																	  // og mellomrom
			}
			//logger.debug("tmpLine: " + tmpLine);
			ukOut.write(tmpLine);
			
			if (i != (this._numTerms - 1)) {
				ukOut.newLine();
			}
		}
		
		ukOut.close();
		
		// skriv skinv til fil
		for (int x = 0; x < this._k; x++) {
			tmpLine = "";
			for (int y = 0; y < this._k; y++) {
				tmpLine += Double.toString(this._Skinv.get(x, y)) + " ";
			}
			
			skinvOut.write(tmpLine);
			
			if (x != (this._k - 1)) {
				skinvOut.newLine();
			}
		}
		
		skinvOut.close();
		
		// skriv vk til fil
		for (int i = 0; i < this._numDocs; i++) { // rad
			tmpLine = "";
			for (int j = 0; j < this._k; j++) { // kolonne
				tmpLine += Double.toString(this._Vk.get(i, j)) + " "; // legger til verdi
																	  // og mellomrom
			}
			//logger.debug("tmpLine: " + tmpLine);
			vkOut.write(tmpLine);
			
			if (i != (this._numDocs - 1)) {
				vkOut.newLine();
			}
		}
		
		vkOut.close();
		
		writeIndexStats();
		System.out.println("writeSVDToFile END");
	}
	
	
		
	/*
	 * Write clusters to files.
	 * docOut tilsvarer vk-matrisen, bare at vektorene er lagt i sine respektive klynger
	 * clusterOut er oversikten over klyngene med sine respektive centroider
	 */
	private void writeIndexStats() throws Exception {
		
		File dest = new File("C:/indexDirectory/luceneIndex/indexInfo/indexStats.txt");
		BufferedWriter indexOut;
		
		indexOut = new BufferedWriter(new FileWriter(dest));
		indexOut.write(Integer.toString(this._numTerms));
		indexOut.newLine();
		indexOut.write(Integer.toString(this._numDocs));
		indexOut.newLine();
		indexOut.write(Integer.toString(this._k));
		indexOut.close();
	}
	
}
