/**
 * @author Jon Rune Paulsen
 * Version: 01.03.2007
 */


package no.omim.engine;

import org.apache.log4j.*;

import org.apache.lucene.index.IndexReader;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.index.TermEnum;
import org.apache.lucene.index.TermFreqVector;
import org.apache.lucene.store.RAMDirectory;
import org.apache.lucene.analysis.Analyzer;

import Jama.*; // For  foreta SVD.

import no.omim.engine.analyzer.OMIMAnalyzer;
import no.omim.engine.analyzer.StopWords;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;

// End import

public class Indexer {
	
	// variables	
	private RAMDirectory _ramDirectory;
	private Analyzer _analyzer;
	private IndexWriter _ramWriter;
	private IndexReader _ramReader;
	
	private Matrix _X;
	private Matrix _U;
	private Matrix _S;
	private Matrix _V;
	private Matrix _Vt;
	private Matrix _Uk;
	private Matrix _Sk;
	private Matrix _Vtk;
	private Matrix _Skinv;
	
	
	private Matrix _XRec;
	
	private int _numTerms;
	private int _numDocs;
	
	private int _numClusters = 5; // antall klynger.
	private ClusterList _arrClusters;
	
	private ArrayList<String> _documents; // TODO sjekk
	//String _cluster[][];
	//double _clusterVector[];
	
	
	private static Logger logger = Logger.getLogger("applogger");
	
	/*
	 * Constructor.
	 */
	public Indexer() throws IOException {
		try {
			this._ramDirectory = new RAMDirectory();
			this._analyzer = new OMIMAnalyzer(StopWords.stopTable);
			this._ramWriter = new IndexWriter(this._ramDirectory, this._analyzer, true);
			this._numTerms = 0;
			this._numDocs = 0;
		}
		catch (IOException ioe)
		{
			throw ioe;
		}
	}
	
	/*
	 * Set up the environment.
	 * @param fileDirectory The directory to the index
	 */
	public void setUp(String fileDirectory) throws IOException {
		try {
			File testData = new File(fileDirectory);
			File[] contents = testData.listFiles();
			
			if (contents != null) {
				for (int i = 0; i < contents.length; i++) {					
					this._ramWriter.addDocument(FileDocument.Document(contents[i]));
					this._ramWriter.optimize();
					this._ramWriter.close();
				}
			}
		}
		catch (IOException ioe) {
			throw ioe;
		}
	}
	
	/*
	 * Transfer to SVD matrix.
	 */
	public void transferToSVDMatrix() throws Exception {
		try {
			this._ramReader = IndexReader.open(_ramDirectory);
			
			TermEnum enumTermer = this._ramReader.terms();
			this._numDocs = this._ramReader.numDocs();
			TermFreqVector[] termFreqVct;
			double normFreq;
			double normWeight;
			double idf;
			int[] arr_ni;
			int k = 0;
			int maxFreq = 0;
			int termFreqInDoc = 0;
			String strTerm;
			String[] arr_termer; // TODO m gjres global og mulighet til  lagre til fil!
			
			// finn antall termer
			while (enumTermer.next())
			{				
				this._numTerms += 1;				
			}
			
			// opprett array over ni
			arr_ni = new int[this._numTerms];
			
			arr_termer = new String[this._numTerms];
			
			// hent ut ni for alle termer
			enumTermer = this._ramReader.terms();
			
			BufferedWriter termWriter;
			File termDestDirectory;
			
			termDestDirectory = new File("indexInfo" + File.separator + "termOverview" + ".txt");
			termWriter = new BufferedWriter(new FileWriter(termDestDirectory));
			
			while (enumTermer.next())
			{
				arr_ni[k] = enumTermer.docFreq();
				arr_termer[k] = enumTermer.term().text();
				
				// lagrer term i fil
				termWriter.write(enumTermer.term().text()); // TODO SJEKK
				
				k += 1;
			}
			
			termWriter.close();
			
			// opprett SVD-matrise
			this._X = new Matrix(this._numTerms, this._numDocs);
			
			// TODO m vel angi strrelsen p this._documents?
			this._documents = new ArrayList<String>();
			
			BufferedWriter docWriter;
			String docInfo;
			File docDestDirectory;
			
			docDestDirectory = new File("indexInfo" + File.separator + "docOverview" + ".txt");
			docWriter = new BufferedWriter(new FileWriter(docDestDirectory));
			
			// for alle dokumenter i index
			for (int i = 0; i < this._numDocs; i++) {
				// returns the stored fields of the ith document in this index.
				// lagrer dokumentstien i en fil
				docInfo = this._ramReader.document(i).get("path");
				docWriter.write(docInfo);
				docWriter.newLine();
				
				// returns the "path" field for the document.
				this._documents.add(docInfo);
				// TODO BLIR IKKE DETTE RIKTIG DA?
				
				
				termFreqVct = this._ramReader.getTermFreqVectors(i);
								
				// finn maks frekvens i dokumentet
				maxFreq = findMaxfrequency(termFreqVct);
				
				// for alle termer i index
				for (int j = 0; j < arr_ni.length; j++) {
					strTerm = arr_termer[j];
					normFreq = 0.0;
					
					// finn termens frekvens i dokumentet
					termFreqInDoc = findTermFrequencyInDocument(strTerm, termFreqVct);					
					
					if (maxFreq > 0)
					{
						normFreq = termFreqInDoc / maxFreq;
					}
					
					idf = Math.log(this._numDocs/arr_ni[j]);
					normWeight = normFreq * idf;
					
					// set(rad, kolonne, verdi)
					this._X.set(j, i, normWeight);
				}
			}
			
			docWriter.close();
		}
		catch (Exception ex) {
			throw ex;
		}
	}
	
	
	
	/*
	 * Finds the max frequency of the term.
	 * @param termFreqVct The term vector to find max frequency of
	 */
	private int findMaxfrequency(TermFreqVector[] termFreqVct) throws Exception
	{
		try
		{
			int max = 0;
			int j;
			int k;
			int[] termFreqs;
			
			for (j = 0; j < termFreqVct.length; j++)
			{
				termFreqs = termFreqVct[j].getTermFrequencies();
				
				for (k = 0; k < termFreqs.length; k++)
				{
					if (termFreqs[k] > max)
					{
						max = termFreqs[k];
					}
				}
			}
			
			return max;
		}
		catch (Exception ex)
		{
			throw ex;
		}
	}
	
	
	
	/*
	 * Finds the term frequency in documents.
	 * @param strTerm The term to find in documents
	 * @param termFreqVct The term vector to search in
	 * @return Frequency
	 */
	private int findTermFrequencyInDocument(String strTerm, TermFreqVector[] termFreqVct)
	{
		int freq = 0;
		int i;
		int j;
		int[] arrFrequencies;
		String[] arrTerms;
		String strTermTemp;
		TermFreqVector termFreqVector;
		
		// alle vektorer
		for (i = 0; i < termFreqVct.length; i++)
		{
			termFreqVector = termFreqVct[i];
			arrTerms = termFreqVector.getTerms();
			arrFrequencies = termFreqVector.getTermFrequencies();
			
			// alle termer i vektor
			for (j = 0; j < arrTerms.length; j++)
			{
				strTermTemp = arrTerms[j];
				logger.debug(arrTerms[j] + " - " + arrFrequencies[j]);
				
				if (strTermTemp.equals(strTerm))
				{
					if (arrFrequencies[j] > freq)
					{
						freq = arrFrequencies[j];
					}
				}
			}		
			
		}
		
		return freq;
	}
	
	
	/*
	 * Perform Singular Value Decomposition and reconstruction of the matrix.
	 */
	public void performSVD() { // TODO sjekk om dette kan vre mten  gjre det p
		performMatrixReduction();
		performRankK();
		reconstructMatrix(); // trenger kanskje ikke  gjre dette (se Garcias tutorial 4
	}
	
	
	
	/*
	 * Perform matrix reduction.
	 */
	private void performMatrixReduction() {
		SingularValueDecomposition SVD = new SingularValueDecomposition(this._X);
		this._U = SVD.getU();
		this._S = SVD.getS();
		this._V = SVD.getV();
		
		this._Vt = _V.transpose(); // Vt (eller D' i Deerwester).
	}
	
	
	
	/*
	 * Perform rank k on matrix U, S, V.
	 */
	private void performRankK() {
		int k = 100;
		
		// Rank k p _U.
		this._Uk = this._U.getMatrix(0, (this._numTerms - 1), 0, (k - 1));
		this._Sk = this._S.getMatrix(0, (k - 1), 0, (k - 1));
		this._Vtk = this._Vt.getMatrix(0, (k - 1), 0, (this._numDocs - 1));
		
		this._Skinv = this._Sk.inverse(); // S invertert
	}
	
	
		
	/*
	 * Reconstruct the original matrix. Trenger trolig ikke denne.
	 */
	private void reconstructMatrix() {
		this._XRec = this._Uk.times(this._Sk.times(this._Vtk)); // Originalmatrise.
	}
	
	
	
	/*
	 * Create clusters.
	 */
	public void createClusters() {
		int size = this._XRec.getRowDimension(); // strrelsen p vektoren
		int i;
		int j;
		Cluster objCluster;
		ClusterList arrClusters = new ClusterList(this._numClusters);	// klyngematrise med 5 celler
		double[] arrClusterVectors; // en klyngevektor med x antall dimensjoner
				
		// lager randomiserte vektorer for klyngene
		for (i = 0; i < this._numClusters; i++) {
			objCluster = new Cluster("cluster" + i); // TODO sjekk
			arrClusterVectors = new double[size];			
			
			for (j = 0; j < size; j++) {				
				if (Math.random() < 0.9) { // br teste "vilkrligheten" til funksjonen
					arrClusterVectors[j] = Math.random();
				}
				else {
					arrClusterVectors[j] = 0 - Math.random();
				}
			}
			
			objCluster.setClusterVectors(arrClusterVectors);
			arrClusters.addCluster(i, objCluster);
		}
	}
	
	
	
	/*
	 * Assign documents to clusters.
	 * 1. Place K points into the space represented by the objects that are being clustered.
	 * These points represent initial group centroids.
	 * 2. Assign each object to the group that has the closest centroid.
	 * When all objects have been assigned, recalculate the positions of the K centroids.
	 * 3. Repeat Steps 2 and 3 until the centroids no longer move.
	 * This produces a separation of the objects into groups from which the metric to be
	 * minimized can be calculated.
	 */
	public void populateClusters() {
		// fyll klyngene med dokumenter.
		// String docs[];
		// String docName = ""; // m fylles med verdi
		boolean repeatPopulation = false;
		Cluster objCluster;
		int size = this._XRec.getRowDimension();
		
		// docs = new String[size / 2]; // array til  holde p dokumentnavn (maks. ca. 6500).
		
		// g gjennom dokumentsamling og legg dokument til klynge.
		// det nedenfor kan gjres p en mer oversiktlig mte ved bruk av funksjoner.
		for (int column = 0; column < this._numDocs; column++) {
			double arrSim[]; // likheten p klyngene og det aktuelle dokumentet.
			double arrDocVec[]; // dokumentvektor.
			
			arrSim = new double[this._numClusters];
			arrDocVec = new double[size];
			
			// lager dokumentvektoren for hvert dokument.
			arrDocVec = createDocVector(size, column);
			
			// for hver klynge, beregn likhet.
			arrSim = computeDocClusterSim(arrDocVec);
			
			// plasser dokument i mest lik klynge.
			objCluster = this._arrClusters.getCluster(closestCluster(arrSim)); // TODO sjekk
			objCluster.setClusterDocuments(this._documents.get(column), column); // TODO sjekk
		}
		
		repeatPopulation = recomputeCentroid();
		
		// hvis sann, repopuler klyngene
		if (repeatPopulation) {
			populateClusters();
		}
	}
	
	
	
	/*
	 * Recompute the centroid of the clusters. 
	 * @return A boolean value whether or not a repopulation of the clusters is needed
	 */
	private boolean recomputeCentroid() { // TODO sjekk
		Cluster objCluster;
		int[] retrievedDocs;
		double[] newCentroid;
		boolean continueRecomp = false;
		
		int clusters = this._numClusters;
		newCentroid = new double[this._numTerms];
		
		for (int i = 0; i < clusters; i++) {
			objCluster = this._arrClusters.getCluster(i);
			retrievedDocs = new int[objCluster.getClusterDocuments().length];
			
			retrievedDocs = objCluster.getClusterDocuments(); // inneholder dokumentnumre
			newCentroid = newCentroid(retrievedDocs); // beregner ny centroidevektor
			
			// TODO hvis ny og gammel centroide ikke er lik, fortsett tilordning av dokumenter 
			if (similarity(newCentroid, objCluster.getClusterVectors()) != 1) { // TODO DENNE M SJEKKES
				continueRecomp = true;
			}
			
			// legger til den nye centroidevektoren
			objCluster.setClusterVectors(newCentroid);
		}
		
		return continueRecomp;
	}
	
	
	
	/*
	 * Creates the new centroid vector.
	 * @param retrievedDocs The numbers for the documents located in the given cluster
	 * @return The new centroid vector
	 */
	private double[] newCentroid(int[] retrievedDocs) { // TODO sjekk
		int retDocs;
		double[] centroid;
		
		retDocs = retrievedDocs.length;
		centroid = new double[this._numTerms]; // TODO evt. _numTerms?
		
		// for hver dimensjon, beregn snitt
		for (int dim = 0; dim < this._numTerms; dim++) {
			double value = 0;
			
			// for hvert dokument, summer verdiene
			for (int doc = 0; doc < retDocs; doc++) {
				value += this._XRec.get(dim, retrievedDocs[doc]); // henter ut verdien fra matrisen  
			}
			
			value /= retDocs; // deler value med antall funnet dokumenter
			centroid[dim] = value; // legger verdien til centroidevektoren
		}
		
		return centroid;
	}
	
	
	
	/*
	 * For each cluster compute document-cluster similarity
	 * @param arrDocVec The vector to compute similarity to clusters with
	 * @return Array of cluster-document similarity
	 */
	private double[] computeDocClusterSim(double[] arrDocVec) {
		double[] docClusterSim;
		Cluster objCluster;
		
		docClusterSim = new double[this._numClusters];
		
		for (int j = 0; j < this._numClusters; j++) {
			objCluster = this._arrClusters.getCluster(j);
			docClusterSim[j] = similarity(objCluster.getClusterVectors(), arrDocVec); // beregner likhet.
		}
		
		return docClusterSim;
	}
	
	
	
	/*
	 * For each document, create document vector
	 * @param vectorSize The size of the vector
	 * @param doc The document to create vector for
	 * @return Document vector
	 * m kunne brukes ved sk, derav public?
	 */
	public double[] createDocVector(int vectorSize, int doc) {
		double[] docVector;
		docVector = new double[vectorSize];
		
		for (int row = 0; row < vectorSize; row++) {
			docVector[row] = this._XRec.get(row, doc);
		}
		
		return docVector;
	}
	
	
	
	/*
	 * Compute similarity.
	 * @param query The query vector
	 * @param doc The document vector
	 * @return The similarity between the query and the document
	 */
	public double similarity(double[] query, double[] doc) {
		int size = this._XRec.getRowDimension();
		double sim;
		double sumDjQ = 0; // summerer dj og q (verste del av formelen).
		double sumDj = 0; // summerer dj (|dj|).
		double sumQ = 0; // summerer q (|q|).
		double sqrtSumDj; // absoluttverdien (absolute value) for dj.
		double sqrtSumQ; // absoluttverdien (absolute value) for q.
		
		for (int term = 0; term < size; term++) {
			double dj = doc[term];
			double q = query[term];
			
			sumDjQ += (dj * q); // verste del av formelen.			
			sumDj += (dj * dj);
			sumQ += (q * q);
		}
		
		sqrtSumDj = Math.sqrt(sumDj); // kvadratroten av dj opphyd i andre (absoluttverdi).
		sqrtSumQ = Math.sqrt(sumQ); // kvadratroten av dj opphyd i andre (absoluttverdi).
		
		sim = sumDjQ / (sqrtSumDj * sqrtSumQ); // cosine similarity
		
		return sim;
	}
	
	
	
	/*
	 * Returns the cluster most similar to the document.
	 * @param arrSim Array with similarities between a document and the clusters
	 * @return The cluster number for the cluster with most similarity to the document
	 */
	private int closestCluster(double[] arrSim) {
		int clusterNumber = 0;
		double maxSimValue = arrSim[0];
		
		for (int i = 1; i < arrSim.length; i++) {
			double arrSimValue = arrSim[i];
			
			if (arrSimValue > maxSimValue) {
				maxSimValue = arrSimValue;
				clusterNumber = i;
			}
		}
		
		return clusterNumber;
	}
	
	
		
	/*
	 * Write clusters to files.
	 */
	public void writeClustersToFile() throws Exception { // TODO sjekk
		Cluster objCluster;
		
		// write cluster path and centroid to file
		File clusterListDirectory;
		BufferedWriter clusterOut;
		
		clusterListDirectory = new File("clusterList" + File.separator + "clusterList.txt");
		clusterOut = new BufferedWriter(new FileWriter(clusterListDirectory));
		
		for (int clusters = 0; clusters < this._numClusters; clusters++) {
			int[] clusterDocuments;
			
			objCluster = this._arrClusters.getCluster(clusters);
			clusterDocuments = new int[objCluster.getClusterDocuments().length];
			clusterDocuments = objCluster.getClusterDocuments();
			
			// write document path/cluster name to file
			File destDirectory;
			BufferedWriter docOut;
			
			// oppretter klyngefil
			destDirectory = new File("clusterList" + clusters + File.separator + objCluster.getClusterID() + ".txt");
			docOut = new BufferedWriter(new FileWriter(destDirectory));
			
			for (int doc = 0; doc < clusterDocuments.length; doc++) {
				// write document name/path to file
				docOut.write(this._ramReader.document(doc).get("path"));
				docOut.newLine();
				String strVector = "[";
				
				for (int docVec = 0; docVec < this._numTerms; docVec++) {
					// create document vector in variable
					String dblToStr = Double.toString(this._XRec.get(docVec, clusterDocuments[doc]));
					
					if (docVec == (this._numTerms - 2)) {
						strVector += dblToStr + "]";
					} else {
						strVector += dblToStr + ";";
					}
				}
				
				// write document vector to file
				docOut.write(strVector);
				docOut.newLine();
			}
			
			// close document writer
			docOut.close();
			
			clusterOut.write(destDirectory.getPath());
			
			String strCentroid = "[";
			double[] tmpCentroid;
			tmpCentroid = new double[this._numTerms];
			tmpCentroid = objCluster.getClusterVectors();
			
			for (int centroidVec = 0; centroidVec < this._numTerms; centroidVec++) {
				// create centroid vector in variable
				String centroidDblToString = Double.toString(tmpCentroid[centroidVec]);
				
				if (centroidVec == (this._numTerms - 2)) {
					strCentroid += centroidDblToString + "]";
				} else {
					strCentroid += centroidDblToString + ";";
				}
			}
			
			// write centroid vector to file
			clusterOut.write(strCentroid);
			clusterOut.newLine();
		}
		
		// close cluster writer
		clusterOut.close();
	}
	
}
