package no.omim.engine;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.math.BigDecimal;
import java.util.ArrayList;

import org.apache.log4j.Logger;

import Jama.Matrix;

public class KMeans {
	protected int _numClusters;
	protected int _numDocs;
	protected int _numTerms;
	protected int _k;
	protected ClusterList _arrClusters;
	protected ArrayList<String> _documents;
	protected Matrix _Vk;
	
	//protected static Logger logger = Logger.getLogger("applogger");
	
	
	/*
	 * Constructor
	 * @param numClusters The number of clusters to be created
	 */
	public KMeans(int numClusters, String dirPath) throws Exception {
		Searcher vkVec = new Searcher(dirPath); // opprinnelig ""
		
		this._numClusters = numClusters;
		
		BufferedReader indexInfoReader;
		indexInfoReader = new BufferedReader(new FileReader(dirPath +
				"/indexInfo/indexStats.txt"));
		
		this._numTerms = Integer.parseInt(indexInfoReader.readLine());
		this._numDocs = Integer.parseInt(indexInfoReader.readLine());
		this._k = Integer.parseInt(indexInfoReader.readLine());
		
		indexInfoReader.close();
		
		BufferedReader vkReader;
		vkReader = new BufferedReader(new FileReader(dirPath + "/SVD/vk.txt"));
		
		String line;
		int i = 0;
		this._Vk = new Matrix(this._numDocs, this._k);
		
		while (null != (line = vkReader.readLine())) {
			double[] tmp = new double[this._k];
			tmp = vkVec.readVector(line);
			
			for (int j = 0; j < this._k; j++) {
				this._Vk.set(i, j, tmp[j]);
			}
			i++;
		}
		
		vkReader.close();
		
		// les dolumentstiene
		BufferedReader docReader;
		docReader = new BufferedReader(new FileReader(dirPath + "/indexInfo/docOverview.txt"));
		
		this._documents = new ArrayList<String>();
		
		while (null != (line = docReader.readLine())) {
			this._documents.add(line);
		}
		
		docReader.close();
	}
	
	
	
	/*
	 * Returns the cluster list containing all of the clusters
	 */
	public ClusterList get() {
		return this._arrClusters;
	}
	
	
	/*
	 * Creates clusters.
	 */
	// legg til parametre
	public void createClusters() {
		int i;
		int j;
		Cluster objCluster;
		
		// klyngematrise med 5 celler
		this._arrClusters = new ClusterList(this._numClusters);
		double[] arrClusterVectors; // en klyngevektor med k antall dimensjoner
		
		// lager randomiserte vektorer for klyngene
		for (i = 0; i < this._numClusters; i++) {
			objCluster = new Cluster("cluster" + i); // TODO sjekk
			arrClusterVectors = new double[this._k];
			//logger.debug("this._k: " + this._k);
			/* TODO HER M JEG GJRE OM!!!
			 * kanskje like greit  plukke ut fem dokumentvektorer som representanter
			 * for klyngene i stedet for  "randomisere" for hver gang
			 * plukk ut flgende dokument:
			 * 141900.txt 1817
			 * 257500.txt 5751
			 * 605923.txt 13336
			 * 300388.txt 6654
			 * 516030.txt 7365
			 * 
			 * siden array og snt starter p 0, s m jeg vel trekke fra 1 p
			 * posisjonen deres.
			 */
			
			// tungvint men n skal utgangscentroidene bli vektorene til de dokumentene
			// som er nevnt ovenfor
			for (j = 0; j < this._k; j++) {
				switch(i) { // DETTE ER KUN TIL TESTING!
					case 0: arrClusterVectors[j] = this._Vk.get(121, j); break;
					case 1: arrClusterVectors[j] = this._Vk.get(7, j); break;
					case 2: arrClusterVectors[j] = this._Vk.get(52, j); break;
				}
				
				// det nedenfor er utgangspunktet
				/*switch(i) {
					case 0: arrClusterVectors[j] = this._Vk.get(1816, j); break;
					case 1: arrClusterVectors[j] = this._Vk.get(5750, j); break;
					case 2: arrClusterVectors[j] = this._Vk.get(6653, j); break;
					case 3: arrClusterVectors[j] = this._Vk.get(7364, j); break;
					case 4: arrClusterVectors[j] = this._Vk.get(13335, j); break;
				}*/
			}
			//logger.debug("Klynge " + objCluster.getClusterID() + " tilordnet");
			objCluster.setClusterVectors(arrClusterVectors);
			this._arrClusters.addCluster(i, objCluster);
			//logger.debug("Klynge " + objCluster.getClusterID() + " lagt til liste");
			//logger.debug("");
		}
		
		//logger.debug("createClusters: END!");
	}
	
	/*
	 * Assign documents to clusters.
	 * 1. Place K points into the space represented by the objects that are being clustered.
	 * These points represent initial group centroids.
	 * 2. Assign each object to the group that has the closest centroid.
	 * When all objects have been assigned, recalculate the positions of the K centroids.
	 * 3. Repeat Steps 2 and 3 until the centroids no longer move.
	 * This produces a separation of the objects into groups from which the metric to be
	 * minimized can be calculated.
	 */
	public void populateClusters() {
		// fyll klyngene med dokumenter.
		boolean repeatPopulation = false;
		Cluster objCluster;
		// g gjennom dokumentsamling og legg dokument til klynge.
		// det nedenfor kan gjres p en mer oversiktlig mte ved bruk av funksjoner.
		
		for (int row = 0; row < this._numDocs; row++) {
			double[] arrSim; // likheten mellom	 klyngene og det aktuelle dokumentet.
			double[] arrDocVec; // dokumentvektor.
			
			arrSim = new double[this._numClusters];
			arrDocVec = new double[this._k];
			// lager dokumentvektoren for hvert dokument.
			arrDocVec = createDocVector(this._k, row); // k = 100, row = hver rad
			// for hver klynge, beregn likhet.
			arrSim = computeDocClusterSim(arrDocVec);
			
			//logger.debug("Plasser dokument i mest lik klynge");
			// plasser dokument i mest lik klynge.
			objCluster = this._arrClusters.getCluster(closestCluster(arrSim)); // TODO sjekk
			//logger.debug("Mest lik klynge: " + objCluster.getClusterID() + "; likhet: " + arrSim[closestCluster(arrSim)]);
			//logger.debug("");
			//logger.debug("row: " + row);
			//logger.debug("Dokumentstrrelse: " + this._documents.size());
			//logger.debug("this._documents.get(row): " + this._documents.get(row));
			objCluster.setClusterDocuments(this._documents.get(row), row); // TODO sjekk
			//logger.debug("Dokumentet er lagt til klyngen");
		}
		//logger.debug("");
		repeatPopulation = recomputeCentroid();
		//logger.debug("repeatPopulation: " + repeatPopulation);
		//logger.debug("");
		
		// hvis sann, repopuler klyngene
		if (repeatPopulation) {
			
			// TEST
			for (int i = 0; i < this._numClusters; i++) {
				objCluster = this._arrClusters.getCluster(i);
				objCluster.clearClusterDocuments();
			}
			// END TEST
			
			populateClusters();
		}
		//logger.debug("populateClusters: END!");
		//logger.debug("");
	}
	
	
	/*
	 * Recompute the centroid of the clusters. 
	 * @return A boolean value whether or not a repopulation of the clusters is needed
	 */
	protected boolean recomputeCentroid() { // TODO sjekk
		Cluster objCluster;
		int[] retrievedDocs;
		double[] newCentroid;
		boolean continueRecomp = false;
		Similarity sim;
		double tmpSim;
		
		int clusters = this._numClusters;
		newCentroid = new double[this._k];
		sim = new Similarity(this._k);
		
		// for hver klynge sjekk om centroiden flytter p seg
		for (int i = 0; i < clusters; i++) {
			objCluster = this._arrClusters.getCluster(i);
			retrievedDocs = new int[objCluster.getClusterDocuments().length];
			
			retrievedDocs = objCluster.getClusterDocuments(); // inneholder dokumentnumre
			newCentroid = newCentroid(retrievedDocs); // beregner ny centroidevektor
			
			/*// TEST
			for (int x = 0; x < this._k; x++) {
				//logger.debug("newCentroid[" + x + "]: " + newCentroid[x]);
				double[] testCentroid = new double[this._k];
				testCentroid = objCluster.getClusterVectors();
				//logger.debug("objCluster[" + x + "]: " + testCentroid[x]);
				//logger.debug("");
			}*/
			//logger.debug("Klyngenummer: " + i);
			for (int y = 0; y < objCluster.getClusterDocuments().length; y++) {
				//logger.debug("Dokument i klyngen: " + retrievedDocs[y]);
			}
			//logger.debug("");
			
			tmpSim = sim.similarity(newCentroid, objCluster.getClusterVectors());
			//logger.debug("tmpSim: " + tmpSim);
			int decimals = 4;
			BigDecimal bd = new BigDecimal(tmpSim);
			bd = bd.setScale(decimals, BigDecimal.ROUND_HALF_UP);
			tmpSim = bd.doubleValue();
			
			//logger.debug("BigDecimal: " + tmpSim);
			
			// TODO hvis ny og gammel centroide ikke er lik, fortsett tilordning av dokumenter 
			if (tmpSim < 1.0) {
				continueRecomp = true;
			}
			
			//logger.debug("continueRecomp: " + continueRecomp);
			//logger.debug("");
			// legger til den nye centroidevektoren
			objCluster.setClusterVectors(newCentroid);
		}
		
		return continueRecomp;
	}
	
	
	/*
	 * Creates the new centroid vector.
	 * @param retrievedDocs The numbers for the documents located in the given cluster
	 * @return The new centroid vector
	 */
	protected double[] newCentroid(int[] retrievedDocs) { // TODO sjekk
		int retDocs;
		double[] centroid;
		
		retDocs = retrievedDocs.length;
		centroid = new double[this._k]; // TODO evt. _numTerms?
		
		// for hver dimensjon, beregn snitt
		for (int dim = 0; dim < this._k; dim++) {
			double value = 0;
			
			// for hvert dokument, summer verdiene
			for (int doc = 0; doc < retDocs; doc++) {
				value += this._Vk.get(retrievedDocs[doc], dim); // henter ut verdien fra matrisen  
			}
			//logger.debug("value: " + value);
			value /= retDocs; // deler value med antall funnet dokumenter
			//logger.debug("value / retDocs: " + value);
			centroid[dim] = value; // legger verdien til centroidevektoren
		}
		
		return centroid;
	}
	
	
	/*
	 * For each document, create document vector
	 * @param vectorSize The size of the vector
	 * @param doc The document to create vector for
	 * @return Document vector
	 * m kunne brukes ved sk, derav public?
	 */
	protected double[] createDocVector(int vectorSize, int doc) {
		double[] docVector;
		docVector = new double[vectorSize];
		
		for (int column = 0; column < vectorSize; column++) {
			docVector[column] = this._Vk.get(doc, column);
		}
		
		return docVector;
	}
	
	
	/*
	 * For each cluster compute document-cluster similarity
	 * @param arrDocVec The vector to compute similarity to clusters with
	 * @return Array of cluster-document similarity
	 */
	protected double[] computeDocClusterSim(double[] arrDocVec) {
		double[] docClusterSim;
		Cluster objCluster;
		Similarity sim; // TODO sjekk
		
		docClusterSim = new double[this._numClusters];
		sim = new Similarity(this._k);
		
		for (int j = 0; j < this._numClusters; j++) {
			objCluster = this._arrClusters.getCluster(j);
			// beregner likhet.
			docClusterSim[j] = sim.similarity(objCluster.getClusterVectors(), arrDocVec);
		}
		
		return docClusterSim;
	}
	
	
	/*
	 * Returns the cluster most similar to the document.
	 * @param arrSim Array with similarities between a document and the clusters
	 * @return The cluster number for the cluster that is the most similar to the document
	 */
	protected int closestCluster(double[] arrSim) {
		int clusterNumber = 0;
		int size = arrSim.length;
		double maxSimValue = arrSim[0];
		
		for (int i = 1; i < size; i++) {
			double arrSimValue = arrSim[i];
			
			if (arrSimValue > maxSimValue) {
				maxSimValue = arrSimValue;
				clusterNumber = i;
			}
		}
		
		return clusterNumber;
	}
	
	
	/*
	 * Writes the k-means clusters to file
	 * @param arrClusters The list of clusters with additional info to be written
	 * @param numClusters The number of clusters to write to file
	 */
	public void writeKMeansClustersToFile(ClusterList arrClusters, int numClusters, String dirPath)
		throws Exception {
		// TODO sjekk
		Cluster objCluster;
		
		// write cluster path and centroid to file
		File clusterListDirectory;
		BufferedWriter clusterOut;
		
		clusterListDirectory = new File(dirPath + "/clusterList/clusterList.txt");
		clusterOut = new BufferedWriter(new FileWriter(clusterListDirectory));
		
		for (int clusters = 0; clusters < numClusters; clusters++) {
			int[] clusterDocuments;
			
			objCluster = arrClusters.getCluster(clusters);
			//clusterDocuments = new int[objCluster.getClusterDocuments().length];
			clusterDocuments = objCluster.getClusterDocuments();
			
			// write document path/cluster name to file
			File destDirectory;
			BufferedWriter docOut;
			
			// oppretter klyngefil
			destDirectory = new File(dirPath + "/clusterList/" + objCluster.getClusterID() + ".txt");
			docOut = new BufferedWriter(new FileWriter(destDirectory));
			
			int clusterDocs = clusterDocuments.length;
			
			for (int doc = 0; doc < clusterDocs; doc++) {
				// write document name/path to file
				String tmp = this._documents.get(clusterDocuments[doc]);
				docOut.write(tmp);
				docOut.newLine();
				String strVector = "";
				
				for (int docVec = 0; docVec < this._k; docVec++) {
					// create document vector in variable
					String dblToStr = Double.toString(this._Vk.get(clusterDocuments[doc], docVec));
					strVector += dblToStr + " ";
				}
				
				// write document vector to file
				docOut.write(strVector);
				
				if (doc != (clusterDocs - 1)) {
					docOut.newLine();
				}
			}
			
			// close document writer
			docOut.close();
			
			clusterOut.write(destDirectory.getPath());
			clusterOut.newLine();
			
			String strCentroid = "";
			double[] tmpCentroid;
			tmpCentroid = new double[this._k];
			tmpCentroid = objCluster.getClusterVectors();
			
			for (int centroidVec = 0; centroidVec < this._k; centroidVec++) {
				// create centroid vector in variable
				String centroidDblToString = Double.toString(tmpCentroid[centroidVec]);
				strCentroid += centroidDblToString + " ";
			}
			
			// write centroid vector to file
			clusterOut.write(strCentroid);
			
			if (clusters != (numClusters - 1)) {
				clusterOut.newLine();
			}
		}
		
		clusterOut.close();
	}
}