package no.omim.engine;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.Reader;
import java.util.ArrayList;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import no.omim.engine.analyzer.OMIMAnalyzer;
import no.omim.engine.analyzer.StopWords;

import org.apache.log4j.Logger;
import org.apache.lucene.analysis.Analyzer;
import org.apache.lucene.queryParser.QueryParser;
import org.apache.lucene.search.Query;

import Jama.Matrix;

public class Searcher {
	
	// Query variables
	private String _srcDirectory;
	private Analyzer _analyzer;
	private Query _query;
	private QueryParser _parser;
	
	// Term variables
	private int _numTerms;
	private int _numDocs;
	private String[] _term;
	private int[] _freq;
	private double[] _normFreq;
	private int _k; // m kunne ordnes slik at den hentes fra et og samme sted (opprinnelig 100)
	
	// File variables
	private ArrayList<String> _path;
	private ArrayList<String> _vector;
	
	private static Logger logger = Logger.getLogger("applogger");
	
	/*
	 * Constructor
	 * @param strSrcDirectory The directory to search
	 */
	public Searcher(String strSrcDirectory) throws IOException {
		this._srcDirectory = strSrcDirectory + "luceneIndex"; // M SETTE DENNE OVER HELE KLASSEN
		logger.debug("strSrcDirectory: " + this._srcDirectory);
		logger.debug("");
		File src;
		BufferedReader readStats;
		
		this._analyzer = new OMIMAnalyzer(StopWords.stopTable);
		this._parser = new QueryParser("query", this._analyzer);
		
		//this._srcDirectory = strSrcDirectory;
		
		src = new File(this._srcDirectory + "/indexInfo/indexStats.txt");
		readStats = new BufferedReader(new FileReader(src));
		
		this._numTerms = Integer.parseInt(readStats.readLine());
		this._numDocs = Integer.parseInt(readStats.readLine());
		this._k = Integer.parseInt(readStats.readLine());
		
		// load clusterlist here
	}
	
	
	/*
	 * Search handler
	 * @param userQuery The query terms given by the user
	 * @return A result list containing the document title, doc path and the similarity
	 */
	public QueryResultList search(String userQuery, boolean blnTwoStep) throws Exception { // twoStepPath brukes kun nr min egen
		String txtTerms;																   // versjon av k-means skal brukes, ellers
		double[] queryVector;															   // er den tom
		double[] convertedQuery;
		double[] simRes;
		Similarity sim;
		QueryResult result;
		String twoStepPath = "/TwoStepKMeans";
		
		logger.debug("Query: " + userQuery);
		
		try {
			this._query = this._parser.parse(userQuery); // TODO M GJRE NOE MED DENNE!!!
		}
		catch (Exception e) {
			logger.error(e);
			throw e;
		}
		
		txtTerms = this._query.toString("query");
		this._term = txtTerms.split("\\s");
		logger.debug("test: " + this._term[0]);
		
		queryVector = new double[this._numTerms];
		
		//normQueryFreq(terms);
		//queryVector = createQueryVector(); // TODO den opprinnelige. like stor som _numTerms
		queryVector = createQueryVec();
		
		// TODO SE GARCIA EQUATION 10, KONVERTERE QUERY TIL DOCUMENT SPACE
		convertedQuery = new double[this._k]; // 100
		
		convertedQuery = convertQuery(queryVector); // m den instansieres frst?
		
		if (blnTwoStep) {
			readFile(this._srcDirectory + twoStepPath + "/clusterList/clusterList.txt");
		} else {
			readFile(this._srcDirectory + "/clusterList/clusterList.txt");		
		}
		
		
		// DETTE ER SAMMENLIGNINGEN AV QUERY OG CLUSTER
		int size = this._path.size(); // skal vre 5
		//int querySize = queryVector.length; // skal vre 100
		
		logger.debug("size: " + size);
		
		sim = new Similarity(this._k); // denne skal vre p 100
		logger.debug("this._k: " + this._k);
		simRes = new double[size]; // strrelsen p det som er innlest (5)
		
		double maxSim = 0;
		int position = 0;
		
		// for hver klynge, sammenlign med query
		for (int i = 0; i < size; i++) {
			double[] cluster;
			
			cluster = new double[this._k];
			cluster = readVector(this._vector.get(i));
			
			simRes[i] = sim.similarity(convertedQuery, cluster);
			
			logger.debug("simRes: " + simRes[i]);
			logger.debug("");
			
			// hvis simRes[i] sin verdi er strre enn maxSim, oppdater maxSim og lagre
			// posisjonen i simRes-listen
			if (maxSim < simRes[i]) {
				maxSim = simRes[i];
				position = i;
			}
		}
		
		if (blnTwoStep) {
			readFile(this._srcDirectory + twoStepPath + "/clusterList/cluster" + position + ".txt");
		} else {
			readFile(this._srcDirectory + "/clusterList/cluster" + position + ".txt");
		}
		
		
		logger.debug("");
		logger.debug("Klyngenummer: " + position);
		logger.debug("");
		
		size = this._path.size(); // skal vre antall dokumenter i den aktuelle klyngen
		//simRes = new double[size];
		
		QueryResultList lstResult;
		lstResult = new QueryResultList(); // tallet skal bort
		lstResult.setClusterID(position + 1);
		
		logger.debug("lstResult opprettet");
		logger.debug("");
		
		// for hvert dokument i klyngen, sammenlign med query
		for (int j = 0; j < size; j++) {
			double[] doc;
			String path;
			int textSize;
			
			doc = new double[this._k];
			doc = readVector(this._vector.get(j));
			
			path = this._path.get(j);
			textSize = path.length();
			
			
			// TODO GJR OM SLIK AT DET OPPRETTES ET NYTT OBJEKT FOR HVERT NYTT RESULTAT
			
			
			
			double tmpSim = sim.similarity(convertedQuery, doc);
			
			//logger.debug("tmpSim: " + String.valueOf(tmpSim));
			
			// hvis likheten er over 0.80 legges den til resultatsettet
			// kan fjernes, men da returneres hele klyngen og det er nok ikke nskelig
			if (tmpSim > 0.80) { // TODO M SJEKKE OM DENNE VERDIEN ER FORNUFTIG
				// legger til resultatobjekt
				result = new QueryResult(path.substring((textSize - 10), (textSize - 4)),	path, tmpSim);
				//result.set(path.substring((textSize - 10), (textSize - 4)),	path, tmpSim);
				
				// legger resultatobjekt til resultatobjektsliste
				lstResult.addResult(j, result);
			}
			//simRes[j] = sim.similarity(convertedQuery, doc);
		}
		
		// returnerer en liste med resultatobjekt som bestr av ID, dokumentsti og likhet
		return lstResult;
	}
	
	
	/*
	 * Converts the given query to document space
	 * @param queryVector The given query vector
	 * @return The converted query vector
	 */
	private double[] convertQuery(double[] queryVector) throws Exception {
		double[] convertedQuery;
		Matrix uk; // rekonstruerer uk-matrisen
		Matrix skinv; // rekonstruerer skinv-matrisen
		Matrix queryT; // transpose
		Matrix conQuery; // midlertidig matrise
		
		//logger.debug("Inni convertQuery");
		
		uk = new Matrix(this._numTerms, this._k);
		//logger.debug("uk: " + uk.getColumnDimension() + " ; " + uk.getRowDimension());
		skinv = new Matrix(this._k, this._k);
		//logger.debug("skinv: " + skinv.getColumnDimension() + " ; " + skinv.getRowDimension());
		queryT = new Matrix(queryVector, 1); // query vector og 1 rad
		//logger.debug("queryT: " + queryT.getColumnDimension() + " ; " + queryT.getRowDimension());
		//logger.debug("test: " + queryT.get(0, 0) + " ; " + queryT.get(0, 1));
		convertedQuery = new double[this._k]; // samme lengde som dokumentene i vk
		
		// insert values into the matrices
		BufferedReader ukReader;
		BufferedReader skinvReader;
		
		ukReader = new BufferedReader(new FileReader(this._srcDirectory + "/SVD/uk.txt"));
		skinvReader = new BufferedReader(new FileReader(this._srcDirectory + "/SVD/skinv.txt"));
		
		String line;
		int x = 0;
		
		while (null != (line = ukReader.readLine())) {
			double[] tmpVec = readVector(line);
			for (int j = 0; j < this._k; j++) {
				uk.set(x, j, tmpVec[j]);
				//logger.debug("tmpVec[" + j + "]: " + tmpVec[j]);
			}
			x++;
		}
		ukReader.close();
		
		x = 0;
		
		while (null != (line = skinvReader.readLine())) {
			double[] tmpVec = readVector(line);
			for (int j = 0; j < this._k; j++) {
				skinv.set(x, j, tmpVec[j]);
				//logger.debug("tmpVecSKINV[" + j + "]: " + tmpVec[j]);
			}
			x++;
		}
		skinvReader.close();
		
		
		
		//logger.debug("test: " + uk.get(0, 0));
		
		conQuery = queryT.times(uk.times(skinv)); // multipliserer som i Garcia eq 10
		
		//logger.debug("");
		//logger.debug("antall kolonner: " + conQuery.getColumnDimension());
		//logger.debug("antall rader: " + conQuery.getRowDimension());
		//logger.debug("");
		
		for (int i = 0; i < this._k; i++) {
			convertedQuery[i] = conQuery.get(0, i); // TODO ER DETTE RIKTIG?
			//logger.debug("convertedQuery[" + i + "]: " + convertedQuery[i]);
		}
		//logger.debug("");
		
		return convertedQuery;
	}
	
	
	/*
	 * Converts the read line to a vector
	 * @param docVector The given read line from a file
	 * @return The vector
	 */
	public double[] readVector(String docVector) {
		double[] vector;
		int vectorSize;
		String[] tmp;
		
		tmp = docVector.split("\\s");
		vectorSize = tmp.length;
		
		//logger.debug("vectorSize: " + String.valueOf(vectorSize));
		
		vector = new double[vectorSize];
		
		for (int i = 0; i < vectorSize; i++) {
			vector[i] = Double.parseDouble(tmp[i]);
			//logger.debug("tmp[" + i + "]: " + tmp[i]);
		}
		
		return vector;
	}
	
	
	/*
	 * Read file. What file to read is decided before this function is called
	 * @param filePath The path to the file to be read
	 */
	private void readFile(String filePath) throws Exception {
		File file = new File(filePath);
		String line;
		int lineNumber = 0;
		BufferedReader fileReader;
		
		fileReader = new BufferedReader(new FileReader(file));
		this._path = new ArrayList<String>();
		this._vector = new ArrayList<String>();
		
		try {
			while (null != (line = fileReader.readLine())) {
				// hvis linjenummeret er et partall
				if (lineNumber % 2 == 0) {
					// les filsti
					//logger.debug("line (path): " + line);
					this._path.add(line);
				} else {
					// les vektor
					//logger.debug("line (vektor): " + line);
					this._vector.add(line); // en celle i listen blir slik: v1 v2 v3... vn
											// hver celle m ha en funksjon som deler opp
											// dimensjonene slik at en fr skt
				}
				
				lineNumber++;
			}
		}
		catch (Exception e) {
			throw e;
		}
	}
	
	
	/*
	 * Create a query vector to compare to the indexed clusters and documents
	 * @param terms The terms given from the user
	 * @return The query vector
	 */
	private double[] createQueryVector() throws Exception {
		double[] queryVector;
		int size = this._term.length;
		
		queryVector = new double[this._numTerms];
		
		// for hver term i query
		for (int i = 0; i < size; i++) {
			int pos = findTermInIndex(this._term[i]);
			logger.debug("POSISJON: " + pos);
			logger.debug("");
			// hvis posisjonen ikke er -1, legg frekvensen inn i posisjonen (default: 1)
			if (pos != -1) {
				queryVector[pos] = 1;
			}
		}
		
		return queryVector;
	}
	
	
	/*
	 * Creates a query vector to compare the indexed clusters and documents
	 * @return The query vector
	 * 
	 * Den som hndterer ArrayList of int-versjonen av findTermInIndex
	 */
	private double[] createQueryVec() throws Exception {
		double[] queryVector;
		int size;
		
		queryVector = new double[this._numTerms];
		
		ArrayList<Integer> pos;
		pos = new ArrayList<Integer>();
		
		pos = findTermsInIndex(this._term);
		size = pos.size();
		
		// for hver posisjon som er funnet
		for (int i = 0; i < size; i++) {
			logger.debug("POSISJON: " + pos.get(i));
			logger.debug("");
			queryVector[pos.get(i)] = 1;
		}
		
		return queryVector;
	}
	
	
	/*
	 * Calculate the normalized query frequencies
	 * TODO lurer p om jeg ikke bare skal drite i normalisert frekvens her og heller
	 * ta utgangspunkt i at brukeren kun skriver inn en term en gang.
	 */
	private void normQueryFreq(String[] queryTerms) {
		int maxFreq;
		ArrayList<String> queryFreq;
		
		int querySize;
		
		queryFreq = new ArrayList<String>();
		
		queryFreq = reduceQuery(queryTerms);
		querySize = queryFreq.size();
		
		this._freq = new int[querySize];
		this._term = new String[querySize];
		
		for (int query = 0; query < querySize; query++) {
			String tmp = queryFreq.get(query);
			String[] splitRes;
			
			splitRes = new String[2];
			splitRes = tmp.split("\\s");
			
			this._term[query] = splitRes[0];
			this._freq[query] = Integer.parseInt(splitRes[1]);
		}
		
		maxFreq = findMaxFreq(); // eks. maxFreq = 2
		
		this._normFreq = new double[querySize];
		
		for (int freqNumber = 0; freqNumber < querySize; freqNumber++) {
			this._normFreq[freqNumber] = 0.5 + (0.5 * this._freq[freqNumber] / maxFreq);
		}
	}
	
	
	/*
	 * Finds the maximum frequency of one term in the query
	 */
	private int findMaxFreq() {
		int maxFreq = this._freq[0];
		int size = this._freq.length;
		
		for (int query = 1; query < size; query++) {
			int tmpFreq;
			
			tmpFreq = this._freq[query];
			
			if (maxFreq < tmpFreq) {
				maxFreq = tmpFreq;
			}
		}
		
		return maxFreq;
	}
	
	
	/*
	 * Reduce the query by decrease the number of the same term to one
	 */
	private ArrayList<String> reduceQuery(String[] queryTerms) {
		ArrayList<String> reducedQuery;
		reducedQuery = new ArrayList<String>();
		int numQueryTerms = queryTerms.length;
		
		for (int term = 0; term < numQueryTerms; term++) {
			String tmpTerm;
			int numTerms = 1;
			tmpTerm = queryTerms[term];
			
			if (!reducedQuery.contains(tmpTerm)) {
				for (int nextTerms = 1; nextTerms < numQueryTerms; nextTerms++) {
					if (queryTerms[term].equals(queryTerms[nextTerms])) {
						numTerms++;
					}
				}
				
				tmpTerm += " " + numTerms;
				reducedQuery.add(tmpTerm);
			}
		}
		
		return reducedQuery;
	}
	
	
	/*
	 * Locate position of query term. If not found in index the term is thrown away.
	 * @param term The query term to be located in the index
	 * @return The position of the located term 
	 */
	private int findTermInIndex(String term) throws Exception {
		int termPos = 0;
		String text;
		boolean found = false;
		
		BufferedReader termReader;
		termReader = new BufferedReader(new FileReader(this._srcDirectory + "/indexInfo/termOverview.txt"));
		
		// m opprette filstrukturen for termsk. Det nedenfor er forslag til metode.
		// TODO Legg til slik at den tar inn en array med termer som sjekkes opp mot hver
		// linje av termReader. M returnere en array med posisjoner.
		
		try {
			while (null != (text = termReader.readLine())) {
				//if (term.equals(text)) {
				if (findSubstring(term.substring(0, (term.length() - 1)) + "?", text)) {
					found = true;
					break;
				}
				termPos++;
			}
			
			if (!found) {
				// hvis termen ikke finnes, returner -1.
				termPos = -1;
			}
		} catch (Exception e) {
			throw e;
		}
		
		return termPos;
	}
	
	
	/*
	 * Locates the position of the query terms
	 * @param terms An array of string containing the query terms
	 * @return An list of term positions
	 * 
	 * ArrayList of int-versjonen av findTermInIndex
	 */
	private ArrayList<Integer> findTermsInIndex(String[] terms) throws Exception {
		ArrayList<Integer> arrTermPos;
		String text;
		
		BufferedReader termReader;
		termReader = new BufferedReader(new FileReader(this._srcDirectory + "/indexInfo/termOverview.txt"));
		arrTermPos = new ArrayList<Integer>();
		
		int termPos = 0;
		
		try {
			while (null != (text = termReader.readLine())) {
				for (int i = 0; i < terms.length; i++) {
					String term = terms[i];
					
					if (findSubstring(term.substring(0, (term.length() - 1)) + "?", text)) {
						//  && !(arrTermPos.contains(termPos)) hvis sjekk p posisjon trengs
						arrTermPos.add(termPos); // legger til termposisjonen
						break; // avslutter for-lkke hvis substring finnes i termReader 
					}
				}
				
				termPos++;
			}
		} catch (Exception e) {
			throw e;
		}
		
		return arrTermPos;
	}
	
	
	/*
	 * Checks whether the given query term is a substring of the index term in the located
	 * position.
	 * @param strPattern The pattern to check with 
	 * @param input The index term to be checked for substring
	 * @return A boolean value based on whether a substring is found or not
	 * 
	 * Denne fungerer, men m vel gjre om p findTermInIndex slik at den returnerer
	 * en array med posisjoner i stedet for bare den frste den finner.
	 */
	private boolean findSubstring(String strPattern, CharSequence input) {
		Pattern pattern = Pattern.compile(strPattern);
		Matcher matcher = pattern.matcher(input);
		
		if (matcher.find()) {
			return true;
		}
		
		return false;
	}

}
