package no.omim.engine;

import org.apache.log4j.Logger;

public class Similarity {
	private int _k;
	
	private static Logger logger = Logger.getLogger("applogger");
	
	public Similarity(int size) {
		this._k = size;
	}
	
	
	/*
	 * Compute similarity.
	 * @param query The query vector
	 * @param doc The document vector
	 * @return The similarity between the query and the document
	 */
	public double similarity(double[] query, double[] doc) {
		double sim;
		double sumDjQ = 0; // summerer dj og q (verste del av formelen).
		double sumDj = 0; // summerer dj (|dj|).
		double sumQ = 0; // summerer q (|q|).
		double sqrtSumDj; // absoluttverdien (absolute value) for dj.
		double sqrtSumQ; // absoluttverdien (absolute value) for q.
		
		for (int term = 0; term < this._k; term++) {
			double dj = doc[term];
			double q = query[term];
			
			sumDjQ += (dj * q); // verste del av formelen.			
			sumDj += (dj * dj);
			sumQ += (q * q);
		}
		
		//logger.debug("sumDjQ: " + String.valueOf(sumDjQ));
		//logger.debug("sumDj: " + String.valueOf(sumDj));
		//logger.debug("sumQ: " + String.valueOf(sumQ));
		
		sqrtSumDj = Math.sqrt(sumDj); // kvadratroten av dj opphyd i andre (absoluttverdi).
		sqrtSumQ = Math.sqrt(sumQ); // kvadratroten av dj opphyd i andre (absoluttverdi).
		
		sim = sumDjQ / (sqrtSumDj * sqrtSumQ); // cosine similarity
		//logger.debug("");
		//logger.debug("sim: " + String.valueOf(sim));
		//logger.debug("");
		return sim;
	}
}
