package no.omim.engine.analyzer;

import no.omim.engine.analyzer.filter.*;
import java.io.*;
import java.util.Set;
import org.apache.log4j.Logger;
import org.apache.lucene.analysis.*;

// An Analyzer builds TokenStreams, which analyze text. It thus represents a
// policy for extracting index terms from text. Typical implementations first
// build a Tokenizer, which breaks the stream of characters from the Reader
// into raw Tokens. One or more TokenFilters may then be applied to the output
// of the Tokenizer. WARNING: You must override one of the methods defined by
// this class in your subclass or the Analyzer will enter an infinite loop.

/**
 * @author Jon Rune Paulsen
 * 
 */
public class OMIMAnalyzer extends Analyzer {
	
	private static Logger logger = Logger.getLogger("applogger");
	private Set stopWords;
	
	public OMIMAnalyzer(String[] stopList) {
		stopWords = StopFilter.makeStopSet(stopList);
	}
    
	/**
	 * Search through every post in the OMIM collection, turns each word into lowercase
	 * and performs Porter stemming, before checking towards the stop table.
	 * @return A stream of tokens.
	 */
    public TokenStream tokenStream(String fieldName, Reader reader) {
    	return new PorterStemFilter(new OMIMStopfilter(new LowerCaseTokenizer(reader), stopWords));
	}
}

